#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "matrix.h"
#include "eigen.h"

/**
 Rutina que lee la matriz de entrada almacenada en un archivo
 Args: nombre_archivo referencia_a_la_matriz_de_salida
**/
MATRIX * lee_archivo( char *nombrearchivo ) 
{
	char linea[80];
	FILE *arch;
	int n, i, j, k;
	double val, **A;
	MATRIX *M;
	char *ptr;

	arch = fopen( nombrearchivo, "r" );
	if( arch == NULL ) {
		fprintf( stderr, "No se puede abrir archivo %s\n", nombrearchivo );
		return( NULL );
	}

	fgets( linea, 80, arch );
	i = sscanf( linea, "%d", &n );
	if ( i!=1 ) {
		fprintf( stderr, "ERROR en la primera lnea del archivo %s\n", nombrearchivo );
		return( NULL );
	}
	// printf( "%d\n", n );
	M = matrix_allocate( n, n, sizeof(double) );
	if ( M == NULL ) {
		fprintf( stderr, "ERROR no hay memoria\n" );
		return( NULL );
	}
	A = (double **)M->ptr;
	for( i=0; i<n; i++ ) {
		fgets( linea, 80, arch );
		ptr = linea;
		for( j=0; j<n; j++ ) {
			k = sscanf( ptr, "%lf", &val );
			if ( k == 1 ) {
				// printf( "%d %d %d\n", i, j, k );
				A[i][j] = val;
				while( *ptr != ' ' && *ptr != '\n' ) ptr++;
				ptr++;
				// printf("%s\n", ptr );
			}
		}
	}
	fclose( arch );
	return( M );
}

/**
	Imprime una matrix A de tamao n x n
**/
void imprime( double **A, int n )
{
	int i, j;

	for( i=0; i<n; i++ ) {
		for( j=0; j<n; j++ ) {
			printf( "%lf ", A[i][j] );
		}
		printf("\n");
	}
	printf("\n");
}


int main( int argc, char **argv )
{
	int i, j;
	char nombre[32];
	MATRIX *M1, *M2, *M3, *M4;
	double **A, **B, **Bcopia, **R;
	int n, cual;
	double eigen[4];
	double *vec, suma;

	if( argc != 2 ) {
		fprintf( stderr, "Args: archivo\n" );
		return( 1 );
	}

	strncpy( nombre, argv[1], 32 );

	M1 = lee_archivo( nombre );
	if( M1==NULL ) {
		fprintf( stderr, "ERROR no hay memoria para la matriz M1\n" );
		return( 2 );
	}

	A = (double **) M1->ptr;

	n = M1->cols;
	imprime( A, n );

	M2 = matrix_allocate( M1->rows, M1->cols, sizeof(double) );
	if( M2==NULL ) {
		fprintf( stderr, "ERROR no hay memoria para la matriz M2\n" );
		matrix_free( M1 );
		return( 3 );
	}

	B = (double **) M2->ptr;
	multiplicaATA( B, A, n );
	imprime( B, n );

	// Hago una copia de la matriz original
	M3 = matrix_allocate( M2->rows, M2->cols + 1, sizeof(double) );
	if( M3==NULL ) {
		fprintf( stderr, "ERROR no hay memoria para la matriz M3\n" );
		matrix_free( M1 );
		matrix_free( M2 );
		return( 4 );
	}
	Bcopia = (double **) M3->ptr;
	for( i=0; i<n; i++ ) {
		for( j=0; j<n; j++ ) {
			Bcopia[i][j] = B[i][j];
		}
	}
	// Memoria para un vector auxiliar de tamao 3*n
	vec = (double *)malloc( 3*n*sizeof(double) );
	if( vec==NULL ) {
		fprintf( stderr, "ERROR no hay memoria para el vector vec\n" );
		matrix_free( M1 );
		matrix_free( M2 );
		matrix_free( M3 );
		return( 5 );
	}
	// Memoria para la matriz R
	M4 = matrix_allocate( M2->rows + 1, M2->cols + 1, sizeof(double) );
	if( M4==NULL ) {
		fprintf( stderr, "ERROR no hay memoria para la matriz M4\n" );
		matrix_free( M1 );
		matrix_free( M2 );
		matrix_free( M3 );
		free(vec);
		return( 6 );
	}
	R = (double **)M4->ptr;

	// El cdigo!
	tridiagonal( B, n, vec );
	// imprime( $B, $n );

	//  Se calculan los tres eigenvalores ms pequenos.
	biseccion( B, n, eigen, 1, 3, vec );
	printf ("Iteraciones: %d\n", (int)eigen[0] );
	for( i=1; i<=3; i++ ) {
		printf ("eigen %d: %lf\n", i, eigen[i] );
	}

	// Selecciona el ms pequeo que sea mayor a 1e+8
	cual = 0;
	for( i=1; i<=3; i++ ) {
		if( fabs(eigen[i]) > 1e-8 ){
			cual = i;
			break;
		}
	}
	if( cual == 0 ) {
		fprintf( stderr, "ERROR: Los eigenvalores seleccionados son todos iguales a cero\n" );
		matrix_free( M1 );
		matrix_free( M2 );
		matrix_free( M3 );
		matrix_free( M4 );
		free(vec);
		exit(2);
	}

	/** Se calcula el eigenvector asociado al eigenvalor encontrado
	    Se usa la matriz original y la descomposicin QR  **/

	printf( "e: %lf\n", eigen[cual] );

	for( i=0; i<n; i++ ) {
		Bcopia[i][n] = 0.5;
		Bcopia[i][i] -= eigen[cual];
	}
	printf( "Hice copia\n" );
	descomposicionQR( Bcopia, n, n+1, R, vec );

	printf( "R:\n" );
	for( i=0; i<n; i++ ) {
		for( j=0; j<=n; j++ ) {
			printf ("%d %d %lf\n", i, j, R[i][j] );
		}
		printf ("\n");
	}
	printf ("\n");

	resuelve( R, n, vec );

	printf ("r: ");
	for( i=0; i<n; i++ ) {
		printf( "%lf ", vec[i] );
	}
	printf("\n");

	suma=0;
	for( i=0; i<n; i++ ) {
		suma += vec[i]*vec[i];
	}
	suma = sqrt( suma );

	printf("rnorm: ");
	for( i=0; i<n; i++ ) {
		vec[i] /= suma;
		printf ("%lf ", vec[i] );
	}
	printf("\n");

	matrix_free( M1 );
	matrix_free( M2 );
	matrix_free( M3 );
	free( vec );

	return 0;
}

